import logging
import time
import datetime
import json
from confluent_kafka import DeserializingConsumer
from confluent_kafka.avro import AvroConsumer, loads
from confluent_kafka.cimpl import TopicPartition
from nubium_utils.custom_exceptions import NoMessageError, ConsumeMessageError, IncompatibleEnvironmentError
from nubium_utils.metrics import MetricsManager
from .confluent_runtime_vars import env_vars
from .confluent_configs import get_kafka_configs, init_consumer_configs, init_transactional_consumer_configs


LOGGER = logging.getLogger(__name__)


def get_consumer(topics, default_schema=None, auto_subscribe=True, cluster=None):
    if isinstance(topics, str):
        topics = topics.split(',')
    LOGGER.debug('Setting up Kafka Consumer')
    consumer = AvroConsumer(
        init_consumer_configs(topics, *get_kafka_configs(), cluster),
        reader_key_schema=loads("""{"type": "string"}"""),
        reader_value_schema=loads(json.dumps(default_schema)) if default_schema else None)
    if auto_subscribe:
        consumer.subscribe(topics)  # in case multiple topics are read from
        LOGGER.info(f'Consumer subscribed to topics:\n{topics}')
    return consumer


def get_transactional_consumer(topics, schema_registry, default_schema=None, auto_subscribe=True, cluster=None):
    if isinstance(topics, str):
        topics = topics.split(',')
    LOGGER.debug('Setting up Kafka Transactional Consumer')
    consumer = DeserializingConsumer(init_transactional_consumer_configs(topics, schema_registry, get_kafka_configs()[0], cluster, default_schema))
    if auto_subscribe:
        consumer.subscribe(topics)  # in case multiple topics are read from
        LOGGER.info(f'Transactional consumer subscribed to topics:\n{topics}')
    return consumer


def _wait_until_message_time(msg_timestamp, guid):
    """
    Wait until the message's timestamp + the deployments offset before handling
    """
    wait_minutes = int(env_vars()['TIMESTAMP_OFFSET_MINUTES'])
    if wait_minutes:
        message_process_time = (msg_timestamp // 1000) + (wait_minutes * 60)

        wait_time = message_process_time - datetime.datetime.timestamp(datetime.datetime.utcnow())

        if wait_time > 0:
            LOGGER.info(f'Waiting {wait_time} seconds before retry message processing continues; GUID {guid}')
            time.sleep(wait_time)


def poll_for_message(consumer, timeout=None):
    """
    Polls the broker for a message using the given timeout.
    If there are no messages to consume, either because None is returned
    or the message error is the no messages error,
    raises a NoMessageError.
    """
    message = consumer.poll(timeout if timeout else int(env_vars()['CONSUMER_POLL_TIMEOUT']))
    if message is None:
        raise NoMessageError
    return message


def handle_consumed_message(message, monitor):
    """
    Handles a consumed message to check for errors, handle retry waits, and log the consumption as a metric

    If the message is returned with a breaking error,
    raises a ConsumeMessageError.

    If the message is valid, waits until the message's timestamp plus
    the current process's time offset before handling the message.
    This allows retry deployments to wait in a non-
    blocking fashion
    """
    guid = None
    try:
        guid = [item[1] for item in message.headers() if item[0] == 'guid'][0].decode()
        if monitor:
            monitor.set_seconds_behind(round(time.time()) - round(message.timestamp()[1]/1000))
        LOGGER.info(f"Message consumed from topic {message.topic()} partition {message.partition()}; GUID {guid}")
        LOGGER.debug(f"Message key: {message.key()}")
    except AttributeError:
        if "object has no attribute 'headers'" in str(message.error()):
            LOGGER.info("Message consumed. No headers, so no guid is available to log.")
    except TypeError:
        LOGGER.info("Message consumed. Headers are None, so no guid is available to log.")
    except IndexError:
        LOGGER.info("Message consumed. Headers found, but no guid is available to log.")

    # If message is None, it can mean that the poll operation timed out,
    # or that there are no more messages to consume

    if message.error():
        if "Broker: No more messages" in str(message.error()):
            LOGGER.warning("Consumer error: %s", message.error())
            raise NoMessageError(message.error())
        else:
            raise ConsumeMessageError(message.error())

    # Wait until message time if using a retry process
    _wait_until_message_time(message.timestamp()[1], guid)

    # Increment the metric for consumed messages by one
    if monitor:
        monitor.inc_messages_consumed(1, message.topic())


def consume_message(consumer, monitor: MetricsManager, timeout=None):
    """
    Consumes a message from the broker while handling errors and waiting if necessary

    If the message is valid, then the message is returned
    """
    message = poll_for_message(consumer, timeout)
    handle_consumed_message(message, monitor)
    return message


def enqueue_auto_commits(consumer, messages):
    """
    A convenience method for staging message(s) to be committed automatically.
    """
    if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT'].lower() == 'false':
        raise IncompatibleEnvironmentError('CONSUMER_ENABLE_AUTO_COMMIT', 'false')
    if isinstance(messages, list):
        for message in messages:
            consumer.store_offsets(message)
    else:
        consumer.store_offsets(messages)


def consume_message_batch(consumer, monitor: MetricsManager, count, timeout=30):
    """
    Consume a batch of messages.

    Requires manual deserialization since consumer.poll() is the method that has deserialization built in on it.
    """
    messages = consumer.consume(num_messages=count, timeout=timeout)
    if not messages:
        raise NoMessageError
    for message in messages:  # attempt avro deserialize
        if message.key() is not None:
            try:
                message.set_key(consumer._serializer.decode_message(message.key(), is_key=True))
            except:
                pass
        if message.value() is not None:
            try:
                message.set_value(consumer._serializer.decode_message(message.value(), is_key=False))
            except:
                pass
        handle_consumed_message(message, monitor)
    return messages


def consume_all_messages(consumer, metric_manager, batch_size=500, min_count_percent=5, timeout=3):
    LOGGER.info("Beginning rapid consumption loop; pulling all messages from the assigned partitions...")
    messages = []
    LOGGER.debug("Seeking to earliest offset...")
    consumer.poll(5)
    partitions = consumer.assignment()
    for partition in partitions:
        consumer.seek(TopicPartition(topic=partition.topic, partition=partition.partition, offset=0))
    while True:
        try:
            new_pull = consume_message_batch(consumer, metric_manager, count=batch_size, timeout=timeout)
            consume_count = len(new_pull)
            messages += new_pull
            LOGGER.debug(f"Added consumed message to queue; current message count: {len(messages)}")
            if consume_count < batch_size * min_count_percent * .01:
                LOGGER.debug(f"Current consume count was too low to continue additional consumes ({consume_count}); stopping consumption...")
                break
        except NoMessageError:
            LOGGER.debug(f"Partitions are fully caught up; stopping consumption...")
            break
    LOGGER.info(f"Rapid consumption loop complete; final message count: {len(messages)}")
    return messages



import logging
from .confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import IncompatibleEnvironmentError, ProducerTimeoutFailure

LOGGER = logging.getLogger(__name__)


def success_headers(headers):
    """
    Updates headers for a message when a process succeeds

    Resets the kafka_retry_count to 0,
    or adds if it didn't originally exist
    """

    try:
        headers = dict(headers)
    except TypeError:
        headers = {}

    headers['kafka_retry_count'] = '0'
    return headers


def produce_message_callback(error, message):
    """
    Logs the returned message from the Broker after producing
    NOTE: Headers not supported on the message for callback for some reason.
    NOTE: Callback requires these args
    """
    LOGGER.debug('Producer Callback...')
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Producer Callback - Message produced successfully')


def consume_message_callback(error, partitions):
    """
    Logs the info returned when a successful commit is performed
    NOTE: Callback requires these args
    """
    LOGGER.debug('Consumer Callback...')
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Consumer Callback - Message consumption committed successfully')


def confirm_produce(producers, attempts=3, timeout=20):
    """
    Ensure that messages are actually produced by forcing synchronous processing. Must manually check the events queue
    and see if it's truly empty since flushing timeouts do not actually raise an exception for some reason.

    NOTE: Only used for synchronous producing, which is dramatically slower than asychnronous.
    """
    if isinstance(producers, dict):
        producers = producers.values()
    else:
        producers = [producers]
    for producer in producers:
        attempt = 1
        while producer.__len__() > 0:
            if attempt <= attempts:
                LOGGER.debug(f"Produce flush attempt: {attempt} of {attempts}")
                producer.flush(timeout=timeout)
                attempt += 1
            else:
                raise ProducerTimeoutFailure


def synchronous_message_handling(producers=None, consumer=None):
    """
    Force the producer/consumer to acknowledge any outstanding, uncommitted messages.
    """
    if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT'].lower() == 'true':
        raise IncompatibleEnvironmentError('CONSUMER_ENABLE_AUTO_COMMIT', 'true')
    if producers:
        LOGGER.debug('Waiting for produce to finish...')
        confirm_produce(producers)
    if consumer:
        LOGGER.debug('Waiting for commits to finalize...')
        consumer.commit(asynchronous=False)


def handle_no_messages(no_msg_exception=None, producers=None):
    """
    Since producer.poll() is typically called within the NU produce_message method, we want to acknowledge
    any outstanding produce attempts while the app has nothing to consume (and thus not actively producing/polling).
    """
    if no_msg_exception:
        LOGGER.debug(no_msg_exception)
    if producers:
        LOGGER.debug('flushing remaining producer queues')
        confirm_produce(producers)


def shutdown_cleanup(producers=None, consumer=None):
    """
    As part of shutdown, make sure all queued up produce messages are flushed, gracefully kill the consumer.
    """
    LOGGER.info('Performing graceful teardown of producer and/or consumer...')
    if consumer:
        LOGGER.debug("Shutting down consumer; no further commits can be queued or finalized.")
        consumer.close()
    if producers:
        LOGGER.debug("Sending/confirming the leftover messages in producer message queue")
        confirm_produce(producers, timeout=30)



import json

from schema_registry.client import SchemaRegistryClient, schema
from schema_registry.serializers import FaustSerializer
from .faust_runtime_vars import env_vars
import logging

LOGGER = logging.getLogger(__name__)

def get_avro_client() -> SchemaRegistryClient:
    """
    Configures a client for the schema registry using an environment variable
    :return: (SchemaRegistryClient)
    """

    LOGGER.info(f'Using the schema server at {env_vars()["SCHEMA_REGISTRY_URL"]}')
    # RHOSAK
    if env_vars()['SCHEMA_REGISTRY_USERNAME']:
        conf = f"https://{env_vars()['SCHEMA_REGISTRY_SERVER']}"
    # LOCAL to Nubium
    elif env_vars()['SCHEMA_REGISTRY_SSL_CA_LOCATION']:
        conf = {
            'url': f"https://{env_vars()['SCHEMA_REGISTRY_SERVER']}",
            'ssl.ca.location': env_vars()['SCHEMA_REGISTRY_SSL_CA_LOCATION'],
        }
    # Nubium to Nubium
    else:
       conf = f"http://{env_vars()['SCHEMA_REGISTRY_SERVER']}"

    return SchemaRegistryClient(conf)


def key_serializer(client) -> FaustSerializer:
    """
    Creates Faust Serializer for generic string schema
    """
    return FaustSerializer(schema_registry_client=client,
                           schema_subject='generic_faust_avro_key',
                           schema=schema.AvroSchema('''{"type":"string"}'''))


def value_serializer(client, topic, schema_dict):
    return FaustSerializer(schema_registry_client=client,
                           schema_subject=topic,
                           schema=schema.AvroSchema(json.dumps(schema_dict)))



"""
confluent-kafka runtime environment variables
"""

from os import environ
from nubium_utils import general_runtime_vars
from nubium_utils.env_var_generator import env_vars_creator


def default_env_vars():
    """
    Environment variables that have defaults if not specified.
    """
    return {
        'CONSUMER_POLL_TIMEOUT': environ.get('CONSUMER_POLL_TIMEOUT', '10'),
        'CONSUMER_TIMEOUT_LIMIT_MINUTES': environ.get('CONSUMER_TIMEOUT_LIMIT_MINUTES', '2'),

        # Performance
        'CONSUMER_ENABLE_AUTO_COMMIT': environ.get('CONSUMER_ENABLE_AUTO_COMMIT', 'true'),
        'CONSUMER_AUTO_OFFSET_RESET': environ.get('CONSUMER_AUTO_OFFSET_RESET', 'latest'),
        'CONSUMER_AUTO_COMMIT_INTERVAL_SECONDS': str(int(environ.get('CONSUMER_AUTO_COMMIT_INTERVAL_SECONDS', '20'))*1000),
        'MESSAGE_BATCH_MAX_MB': str(int(float(environ.get('MESSAGE_BATCH_MAX_MB', '2')) * 2 ** 20)),
        'MESSAGE_TOTAL_MAX_MB': str(int(float(environ.get('MESSAGE_TOTAL_MAX_MB', '3')) * 2 ** 20)),
        'MESSAGE_QUEUE_MAX_MB': str(int(float(environ.get('MESSAGE_QUEUE_MAX_MB', '10')) * 2 ** 10)),

        # All required below if using "retry" logic
        'CONSUME_TOPICS': environ.get('CONSUME_TOPICS', ''),
        'TIMESTAMP_OFFSET_MINUTES': environ.get('TIMESTAMP_OFFSET_MINUTES', '0'),
        'RETRY_COUNT_MAX': environ.get('RETRY_COUNT_MAX', '0'),
        'PRODUCE_RETRY_TOPICS': environ.get('PRODUCE_RETRY_TOPICS', ''),
        'PRODUCE_FAILURE_TOPICS': environ.get('PRODUCE_FAILURE_TOPICS', '')}


def required_env_vars():
    """
    Environment variables that require a value (aka no default specified).
    """
    return {}


def all_env_vars():
    return {
        **general_runtime_vars.all_env_vars(),
        **default_env_vars(),
        **required_env_vars()
    }


env_vars = env_vars_creator(all_env_vars)



from .instrumented_app import InstrumentedApp
from .app_wrapper import FaustAppWrapper, FaustMessage
from .helpers import get_config, get_ssl_context
from .avro_utils import get_avro_client, key_serializer, value_serializer
from .kiwf_health_check import kiwf_log_analysis



from unittest import mock, TestCase
from unittest.mock import patch
from nubium_utils.faust_utils import (
    FaustAppWrapper,
    InstrumentedApp,
    get_ssl_context,
    faust_runtime_vars
)
from nubium_utils.faust_utils.helpers import SASLCredentials
import importlib


class ImplementedAppWrapper(FaustAppWrapper):

    def _init_agents(self):
        pass

    def _init_records(self):
        pass

    def _init_serializers(self):
        pass

    def _init_tables(self):
        pass

    def _init_topics(self):
        pass


class TestFaustAppWrapper(TestCase):
    """
    Tests the app wrapper
    """
    def tearDown(self):
        self.env_patch.stop()

    def setUp(self):
        self.env_patch = patch.dict(
            'os.environ',
            {
                "HOSTNAME": "TEST",
                "APP_NAME": "TEST",
                "TOPIC_PARTITIONS": "1",
                "MP_PROJECT": "none",
                "BROKER_TOPIC_NAME": 'test_topic',
                "SCHEMA_REGISTRY_URL": "test_registry:8081"
            }
        )
        self.env_patch.start()
        importlib.reload(faust_runtime_vars)

    def test_app_creation(self):

        test_faust_config = {
            'id': 'test-id',
            'broker': 'kafka://test',
            'store': "memory://"
        }
        mock_monitor = mock.MagicMock()

        app_wrapper = ImplementedAppWrapper(test_faust_config, avro_client=None, metrics_manager=mock_monitor)
        assert app_wrapper.app.__class__ == InstrumentedApp

    def test_add_input_topic(self):

        test_faust_config = {
            'id': 'test-id',
            'broker': 'kafka://test',
            'store': "memory://"
        }
        mock_monitor = mock.MagicMock()
        mock_serializer = mock.MagicMock()

        app_wrapper = ImplementedAppWrapper(test_faust_config, avro_client=None, metrics_manager=mock_monitor)
        faust_topic = app_wrapper.add_input_topic('test_topic', mock_serializer)

        # Assert that method returns a faust_topic
        assert faust_topic.get_topic_name() == 'test_topic_Internal_test-id'

        # Assert that a input_topic is added
        input_topic = app_wrapper.input_topics['test_topic']

        # Assert that a converter agent was added
        converter_agent = app_wrapper.app.agents.data['test_topic_agent_internal']
        assert converter_agent.channel == input_topic


class TestFaustHelpers(TestCase):

    def test_no_ssl_context(self):
        """
        get_ssl_context returns None when `USE_SSL` env variable is False
        """
        with mock.patch('nubium_utils.faust_utils.helpers.env_vars') as mock_env:
            mock_env.return_value = {'RHOSAK_USERNAME': ''}
            output = get_ssl_context()
            assert output == {}

    def test_get_ssl_context(self):
        """
        SSL context configures from env variables when `USE_SSL` is True
        """
        with mock.patch('nubium_utils.faust_utils.helpers.env_vars') as mock_env:
            mock_env.return_value = {
                'RHOSAK_USERNAME': 'test_user',
                'RHOSAK_PASSWORD': 'test_pass'}
            assert isinstance(get_ssl_context()['broker_credentials'], SASLCredentials)



from .dude_utils import dude_delete_topics, dude_produce_input, dude_consume_output
from .setup_teardown_utils import setup_teardown_env_vars, setup_teardown_app, setup_teardown_test_data
from .validation_utils import actual_results, expected_results, validate_results



from pathlib import Path
import rocksdb
from os import environ


class RDB:
    def __init__(self, table_name):
        opts = rocksdb.Options()
        opts.create_if_missing = True
        opts.max_open_files = 300000
        opts.write_buffer_size = 67108864
        opts.max_write_buffer_number = 3
        opts.target_file_size_base = 67108864

        opts.table_factory = rocksdb.BlockBasedTableFactory(
            filter_policy=rocksdb.BloomFilterPolicy(10),
            block_cache=rocksdb.LRUCache(2 * (1024 ** 3)),
            block_cache_compressed=rocksdb.LRUCache(500 * (1024 ** 2)))

        self.rdb = rocksdb.DB(f"{environ['TABLE_PATH']}/{table_name}.rocksdb", opts)

    def write(self, key, value):
        self.rdb.put(key.encode(), value.encode())

    def write_batch(self, write_dict):
        batch = rocksdb.WriteBatch()
        for k, v in write_dict.items():
            try:
                v = v.encode()
            except ValueError:
                pass
            batch.put(k.encode(), v)
        self.rdb.write(batch)

    def read(self, k):
        result = self.rdb.get(k.encode())
        if result:
            return result.decode()
        return result

    def delete(self, key):
        self.rdb.delete(key.encode())

    def close(self):
        # TODO: Add flush method
        self.rdb.close()



import itertools
from time import sleep

from confluent_kafka.cimpl import KafkaException
from confluent_kafka.error import KafkaError
from confluent_kafka.admin import AdminClient, NewTopic
from fastavro.validation import validate
from unittest.mock import MagicMock
from nubium_utils import log_and_raise_error
from nubium_utils.confluent_utils import get_producers, produce_message, get_consumer, enqueue_auto_commits, shutdown_cleanup, handle_no_messages
from nubium_topic_configs.topics import primary_topics, internal_topics, changelog_topics
from nubium_topic_configs.clusters import cluster_topic_map
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.confluent_utils.consumer_utils import consume_all_messages
from nubium_utils.custom_exceptions import NoMessageError, SignalRaise


class KafkaToolbox:
    """
    Helpful functions for interacting with Kafka (mainly RHOSAK).
    Allows you to interface with multiple clusters at once by default assuming they are defined via your environment
    variables NUBIUM_CLUSTER_{N}. Actions include:

    Topics (multiple): create, delete, list.

    Messages (multiple): produce
    """

    def __init__(self, config=None):
        self.admin_clients = {}
        self.producers = {}
        self.consumer = {}
        self.default_cluster_config = {}
        self.nubium_topics = {**primary_topics, **internal_topics, **changelog_topics}
        self.cluster_topic_map = cluster_topic_map()
        self.nubium_clusters = [c for c in set(self.cluster_topic_map.values()) if c]
        self.nubium_primary_topic_clusters = {
            cluster: [topic for topic in primary_topics if self.cluster_topic_map[topic] == cluster]
            for cluster in self.nubium_clusters}
        if not config and env_vars().get('TEST_CLUSTER'):
            config = {'bootstrap.servers': env_vars()['TEST_CLUSTER']}
        if config:
            self.default_cluster_config = config
            # self.add_admin_cluster_client(additional_configs=config, custom_server_name='default')

    def add_admin_cluster_client(self, server_url='', additional_configs=None, custom_server_name=None):
        """
        Add a new client connection for a given cluster via the 'self.admin_client' dict.

        By default, uses your environment to establish the client; usually all you need to provide is 'server_url'.
        Any 'additional_configs' will override any environment-based defaults.

        You can also provide your own client nickname via 'custom_server_name' (mostly used to define the
        "default" server at KafkaToolbox init time), else defaults to 'server_url'.
        """
        if self.admin_clients.get(custom_server_name if custom_server_name else server_url) is None:
            sleep(.2)
            config = {
                "bootstrap.servers": server_url
            }
            if 'localhost' not in server_url:
                config.update({
                    "sasl.username": env_vars().get("RHOSAK_USERNAME"),
                    "sasl.password": env_vars().get("RHOSAK_PASSWORD"),
                    "security.protocol": "sasl_ssl",
                    "sasl.mechanisms": "PLAIN",
                })
            if additional_configs:
                config.update(additional_configs)
            client = AdminClient(config)
            self.admin_clients[config['bootstrap.servers']] = client
            if custom_server_name:
                self.admin_clients[custom_server_name] = client

    def _add_default_client(self):
        if self.default_cluster_config and not self.admin_clients.get('default'):
            self.add_admin_cluster_client(
                server_url=self.default_cluster_config.get('bootstrap.servers', ''),
                additional_configs=self.default_cluster_config,
                custom_server_name='default')
        else:
            raise ValueError(f'No "default" broker configs were provided; please add a valid "default" broker config.'
                             ' This is best done by remaking the KafkaToolbox object with an appropriate "TEST_CLUSTER"'
                             'env var url, or manually defining a confluent broker config via the "config" argument.')

    def create_topics(self, topic_cluster_dict, num_partitions=3, replication_factor=3, topic_config=None,
                      ignore_nubium_topic_cluster_maps=False):
        """
        Accepts a "topic_cluster_dict" like: {'topic_a': 'cluster_url_x', 'topic_b': 'cluster_url_y', 'topic_c': ''}
        If cluster url is provided, will default to that; else, will attempt to derive using nubium_schemas
        if "use_nubium_topic_cluster_maps" is True.
        Otherwise, will default to provided "cluster" arg (which by default, is set to the admin_client you optionally
        instantiated the KafkaToolbox object with).

        Separately, if "ignore_nubium_topic_cluster_maps" is False, creates topic with predefined nubium_schemas configs.
        Else, will use the args "num_partitions", "replication_factor", "topic_config".
        """
        topic_dict = {**{clust: [] for clust in self.nubium_clusters}, 'default': []}
        self._add_default_client()
        for topic, clust in topic_cluster_dict.items():
            if topic not in primary_topics or ignore_nubium_topic_cluster_maps:
                topic_dict[clust if clust else 'default'].append(
                    NewTopic(
                        topic=topic,
                        num_partitions=num_partitions,
                        replication_factor=replication_factor,
                        config=topic_config if topic_config else {}
                    )
                )
            else:
                topic_dict[clust if clust else self.cluster_topic_map[topic]].append(
                    NewTopic(
                        topic=topic,
                        num_partitions=primary_topics[topic].partitions,
                        replication_factor=primary_topics[topic].replication_factor,
                        config=primary_topics[topic].config
                    )
                )
        for clust, topics in topic_dict.items():
            if topics:
                self.add_admin_cluster_client(server_url=clust)
                for topic in topics:
                    try:
                        _wait_on_futures_map(self.admin_clients[clust].create_topics([topic]))
                        print(f"Topic created: {topic}")
                    except KafkaException as e:
                        if e.args[0].code() == KafkaError.TOPIC_ALREADY_EXISTS:
                            print(f"Topic already exists: {topic}")
                            pass

    def delete_topics(self, topic_cluster_dict, ignore_nubium_topic_cluster_maps=False):
        """
        Accepts a "topic_cluster_dict" like: {'topic_a': 'cluster_url_x', 'topic_b': 'cluster_url_y', 'topic_c': ''}
        If cluster url is provided, will default to that; else, will attempt to derive using nubium_schemas
        if "ignore_nubium_topic_cluster_maps" is False.
        Otherwise, will default to provided "cluster" arg (which by default, is set to the admin config you optionally
        instantiated the KafkaToolbox object with).
        """
        topic_dict = {**{clust: [] for clust in self.nubium_clusters}, 'default': []}
        self._add_default_client()
        for topic, clust in topic_cluster_dict.items():
            if topic not in primary_topics or ignore_nubium_topic_cluster_maps:
                topic_dict[clust if clust else 'default'].append(topic)
            else:
                topic_dict[clust if clust else self.cluster_topic_map[topic]].append(topic)
        for clust, topics in topic_dict.items():
            if topics:
                self.add_admin_cluster_client(server_url=clust)
                for topic in topics:
                    try:
                        _wait_on_futures_map(self.admin_clients[clust].delete_topics([topic]))
                        print(f"Topic deleted: {topic}")
                    except KafkaException as e:
                        if e.args[0].code() == KafkaError.UNKNOWN_TOPIC_OR_PART:
                            print(f"Topic deletion failed (likely because it didn't exist): {topic}")
                            pass

    def list_topics(self, by_topic=False, all_clusters=True, mirrors=False, cluster=''):
        """
        Allows you to list all topics across all cluster instances.

        If you want to return the output in a format that is used by the other toolbox functions, then
        you can change "by_topic"=True; the default format makes it easier to read.

        If you disable all_clusters, will default to to the admin config you optionally
        instantiated the KafkaToolbox object with.

        Will not include the mirrored topics by default, but can toggle with "mirrors".
        """
        def _mirror_include(topic):
            if topic.startswith('nubium_'):
                return mirrors
            return True

        def _valid(topic):
            return not topic.startswith('__') and 'schema' not in topic

        self._add_default_client()
        clients = [self.admin_clients['default']]
        if cluster:
            self.add_admin_cluster_client(server_url=cluster)
            clients = [self.admin_clients[cluster]]
        if all_clusters:
            for clust in self.nubium_clusters:
                self.add_admin_cluster_client(server_url=clust)
                clients += [self.admin_clients[clust]]
        client_server_map = {client: [s for s, c in self.admin_clients.items() if c == client][0] for client in clients}
        by_cluster = {clust: [topic for topic in list(client.list_topics().topics) if _valid(topic) and _mirror_include(topic)]
                      for client, clust in client_server_map.items()}
        if by_topic:
            topics = {topic: clust for clust, topics in by_cluster.items() for topic in topics}
            return {k: topics[k] for k in sorted(topics.keys())}
        return {k: sorted(by_cluster[k]) for k in sorted(by_cluster.keys())}

    def produce_messages(self, topic, message_list, schema=None, cluster=None):
        """
        Produce a list of messages (each of which is a dict with entries 'key', 'value', 'headers') to a topic.

        Must provide the schema if you haven't produced to that topic on this instance, or if using dude CLI.
        """
        # TODO extra producer error handling?
        if not self.producers.get(topic):
            if not schema:
                raise ValueError('Schema not provided and the topic producer instance was not previously configured. '
                                 'Please provide the schema!')
            if topic not in self.nubium_topics and not cluster:
                cluster = self.default_cluster_config['bootstrap.servers']
            self.producers.update(get_producers({topic: schema}, return_as_singular=False, cluster=cluster))
        # validate(message, schema)  # Not sure we need this?
        producer = self.producers[topic]
        for message in message_list:
            produce_message(
                producer=producer,
                producer_kwargs=message,
                metrics_manager=MagicMock())  # TODO: change NU to allow a None here? dunno
        producer.flush()

    def consume_messages(self, topic, cluster=None):
        """
        Consume a list of messages (each of which is a dict with entries 'key', 'value', 'headers') from a topic.
        """
        if not self.consumer.get(topic):
            if topic not in self.nubium_topics and not cluster:
                cluster = self.default_cluster_config['bootstrap.servers']
            self.consumer = get_consumer(topic, cluster=cluster)

        try:
            messages = consume_all_messages(consumer=self.consumer, metric_manager=MagicMock())
            enqueue_auto_commits(self.consumer, messages)
            shutdown_cleanup(consumer=self.consumer)
            return messages

        except NoMessageError as e:
            handle_no_messages(e)

        except SignalRaise:
            raise

        except Exception as error:
            log_and_raise_error(MagicMock(), error)


    # def wipe_topic(self, topic):
    #     topics = self.admin_clients[''].list_topics().topics
    #     topic_to_recreate = topics[topic]
    #     partitions = topic_to_recreate.partitions.values()
    #     replicas = set(itertools.chain.from_iterable(p.replicas for p in partitions))
    #     self.admin_clients[''].delete_topic(topic)
    #     while True:
    #         try:
    #             # until KIP-516 is implemented, success of topic creation seems
    #             # to be the only way to detect topic deletion is finished.
    #             self.admin_clients[''].create_topics(topic, num_partitions=len(partitions), replication_factor=len(replicas))
    #             break
    #         except:  # pylint: disable=bare-except
    #             sleep(0.1)


def _wait_on_futures_map(futures):
    for future in futures.values():
        future.result()
        assert future.done()
        sleep(.1)



import logging
import subprocess
from os import environ
from time import sleep

import pytest

LOGGER = logging.getLogger(__name__)


@pytest.fixture()
def dude_delete_topics():
    LOGGER.info(f"Deleting topics to start fresh...\n[{environ['INPUT_TOPIC']},{environ['OUTPUT_TOPIC']}]")
    subprocess.run(f"dude topics delete --topics {environ['INPUT_TOPIC']},{environ['OUTPUT_TOPIC']} --ignore-cluster-maps", shell=True)


@pytest.fixture()
def dude_produce_input():
    LOGGER.info("Producing test messages to the input topic...")
    subprocess.run(f"dude topics produce --topic {environ['INPUT_TOPIC']} --message-file {environ['TEST_DATA_IN']} --schema {environ['TEST_DATA_IN_SCHEMA']}", shell=True)
    sleep(30)  # wait for app to process the messages


@pytest.fixture()
def dude_consume_output():
    LOGGER.info("Consuming test messages from the output topic and dumping them to a file...")
    subprocess.run(f"dude topics consume --topic {environ['OUTPUT_TOPIC']} --message-file {environ['TEST_DATA_OUT']}", shell=True)



from .consumer_utils import get_transactional_consumer, consume_message, get_consumer
from .message_utils import shutdown_cleanup
from .producer_utils import get_transactional_producer, produce_message
from nubium_utils.general_utils import parse_headers
from confluent_kafka import TopicPartition
from nubium_utils.confluent_utils.confluent_configs import init_schema_registry_configs
from nubium_utils.custom_exceptions import NoMessageError
import logging
import sys
# from orjson import dumps, loads
# changelog_schema = {"type": "bytes"}
from json import dumps, loads
changelog_schema = {"type": "string"}
from random import random

from copy import deepcopy
from nubium_utils.confluent_utils.rocksb_utils import RDB
from os import environ
from time import sleep

LOGGER = logging.getLogger(__name__)

class RunTableRecovery(Exception):
    def __init__(self):
        pass

class Transaction:
    def __init__(self, message, producer, consumer, metrics_manager):
        self.producer = producer
        self.consumer = consumer
        self.metrics_manager = metrics_manager
        self.message = message
        
        self._committed = False
        self._active_transaction = False

    def key(self):
        return self.message.key()

    def value(self):
        return self.message.value()

    def headers(self):
        return parse_headers(self.message.headers())

    def topic(self):
        return self.message.topic()

    def partition(self):
        return self.message.partition()

    def offset(self):
        return self.message.offset()

    def produce(self, producer_kwargs):
        self.producer.poll(0)
        if not self._active_transaction:
            self.producer.begin_transaction()
            self._active_transaction = True
        produce_message(self.producer, producer_kwargs, self.metrics_manager, self.headers())
        self.producer.poll(0)

    def commit(self, mark_committed=True):
        if self._active_transaction:
            self.producer.send_offsets_to_transaction(
                [TopicPartition(self.topic(), self.partition(), self.offset() + 1)], self.consumer.consumer_group_metadata())
            self.producer.commit_transaction()
            self.producer.poll(0)
            self._committed = mark_committed


class TableTransaction(Transaction):
    def __init__(self, message, producer, consumer, metrics_manager, changelog_topic):
        self.message = message
        self.changelog_topic = changelog_topic

        self._changelog_updated = False
        self._pending_table_write = None

        super().__init__(self.message, producer, consumer, metrics_manager)

    def read_table_entry(self):
        return self._rdb_read()
        
    def update_table_entry(self, value):
        self._pending_table_write = deepcopy(value)
        if isinstance(self._pending_table_write, (list, dict)):
            LOGGER.debug(f'attempting json dumps of {self._pending_table_write}')
            self._pending_table_write = dumps(self._pending_table_write)

    def delete_table_entry(self):
        self._pending_table_write = '-DELETED-'
        
    def _update_changelog(self):
        self.produce(dict(
            topic=self.changelog_topic,
            key=self.key(),
            value=self._pending_table_write
        ))
        self._changelog_updated = True
        self.producer.poll(0)

    def _recover_table_via_changelog(self):
        value = self.value()
        try:
            value = loads(value)
        except:
            pass
        if value == '-DELETED-':
            self.delete_table_entry()
        else:
            self.update_table_entry(value)
        super().commit()
        self._rdb_write(self._pending_table_write)

    def _rdb_write(self, value):
        if self._pending_table_write == '-DELETED-':
            LOGGER.debug('Finalizing table entry delete...')
            rdb_tables[self.partition()].delete(self.key())
        else:
            LOGGER.debug(f'Finalizing table entry write:\npartition{self.partition()},\nkey:{self.key()},\nvalue:{self._pending_table_write}...')
            rdb_tables[self.partition()].write_batch(
                {self.key(): self._pending_table_write,
                 'offset': str(self._rdb_offset() + 2)})
        self._pending_table_write = None

    def _rdb_read(self):
        value = rdb_tables[self.partition()].read(self.key())
        try:
            value = loads(value)
        except:
            pass
        LOGGER.debug(f'Read table value: {value}')
        return value

    def _rdb_offset(self):
        value = rdb_tables[self.partition()].read('offset')
        if not value:
            value = self.offset() if self.offset() else 0
        return int(value)
    
    def commit(self):
        if not self._changelog_updated and self._pending_table_write:
            self._update_changelog()
        super().commit(mark_committed=False)
        if self._pending_table_write:
            self._rdb_write(self._pending_table_write)


class TransactionApp:
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict, 
                 metrics_manager=None, schema_registry=None, cluster=None, consumer=None, producer=None):
        self.app_function = app_function
        self.metrics_manager = metrics_manager
        if not schema_registry:
            self.schema_registry = init_schema_registry_configs(as_registry_object=True)
        if not consumer:
            self.consumer = get_transactional_consumer(consume_topics_list, self.schema_registry, cluster=cluster)
        if not producer:
            self.producer = get_transactional_producer(produce_topic_schema_dict, self.schema_registry, cluster=cluster)

    def consume(self, timeout=None):
        return consume_message(self.consumer, self.metrics_manager, timeout)

    def run(self, *args, **kwargs):
        try:
            while True:
                transaction = None
                try:
                    transaction = Transaction(self.consume(), self.producer, self.consumer, self.metrics_manager)
                    self.app_function(transaction, *args, **kwargs)
                    if not transaction._committed:
                        transaction.commit()
                except NoMessageError:
                    self.producer.poll(0)
                except Exception as e:
                    raise
        finally:
            if transaction:
                if transaction._active_transaction:
                    self.producer.abort_transaction(10)
            shutdown_cleanup(consumer=self.consumer)


class TableApp(TransactionApp):
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict, 
                 metrics_manager=None, schema_registry=None, cluster=None, consumer=None, producer=None):
        self.changelog_topic = f"{environ['APP_NAME']}__changelog"
        if self.changelog_topic not in produce_topic_schema_dict:
            produce_topic_schema_dict.update({self.changelog_topic: changelog_schema})
        global rdb_tables
        rdb_tables = {}
        if not schema_registry:
            self.schema_registry = init_schema_registry_configs(as_registry_object=True)
        if not consumer:
            self.consumer = self._set_table_consumer(consume_topics_list, self.schema_registry, cluster=cluster)
        super().__init__(
            app_function, consume_topics_list, produce_topic_schema_dict,
            metrics_manager=metrics_manager, schema_registry=schema_registry, cluster=cluster, consumer=self.consumer, producer=producer)

    def _set_table_consumer(self, topic, schema_registry, default_schema=None, cluster=None):
        consumer = get_transactional_consumer(topic, schema_registry, default_schema, False, cluster)
        setattr(consumer, 'topic', topic)
        setattr(consumer, 'changelog_topic', self.changelog_topic)
        setattr(consumer, 'table_recovery_status', {})
        setattr(consumer, 'temp_assign_partitions', [])
        setattr(consumer, 'temp_assign_recovery_partitions', {})
        consumer.subscribe([topic], on_assign=_rocksdb_assign, on_revoke=_rocksdb_unassign, on_lost=_rocksdb_unassign)
        LOGGER.debug('Table consumer initialized')
        return consumer
    
    def _check_changelog_offset(self):
        self.consumer.position()
    
    def _table_recovery(self):
        checks_left = 2
        # TODO: This while loop may not be necc anymore since I'm doing the exception approach now
        while not self.consumer.table_recovery_status or checks_left:
            try:
                # Must poll with consumer to ensure all data on the consumer object is properly populated
                transaction = TableTransaction(self.consume(timeout=2), self.producer, self.consumer,
                                               self.metrics_manager, self.changelog_topic)
                # Once the consumer is appropriately populated...
                checks_left = 0
            except NoMessageError:
                LOGGER.debug('Still waiting for table transaction status...')
                checks_left -= 1
        LOGGER.debug(f'table_recovery_status before recovery attempt: {self.consumer.table_recovery_status}')
        LOGGER.debug(f'temp_assign_recovery_partitions before recovery attempt: {self.consumer.temp_assign_recovery_partitions}')
        seek_list = [p_obj for p, p_obj in self.consumer.temp_assign_recovery_partitions.items() if p in self.consumer.table_recovery_status]
        for p_obj in seek_list:
            self.consumer.seek(p_obj)
        if self.consumer.table_recovery_status:
            LOGGER.info('BEGINNING TABLE RECOVERY PROCEDURE')
            checks_left = 2
            while (checks_left and self.consumer.table_recovery_status):
                try:
                    transaction = TableTransaction(self.consume(timeout=3), self.producer, self.consumer, self.metrics_manager,
                                                   self.changelog_topic)
                    LOGGER.debug(f'Recovery write is {transaction.value()}')
                    transaction._recover_table_via_changelog()
                    p = transaction.partition()
                    LOGGER.debug(f"transaction_offset - {transaction.offset() + 2}, watermark - {self.consumer.table_recovery_status[p]['watermarks'][1]}")
                    if self.consumer.table_recovery_status[p]['watermarks'][1] - (transaction.offset() + 2) <= 0:
                        LOGGER.info(f'table partition {p} fully recovered!')
                        del self.consumer.table_recovery_status[p]
                except NoMessageError:
                    checks_left -= 1
            unassign = {p: p_obj for p, p_obj in self.consumer.temp_assign_recovery_partitions.items() if p not in self.consumer.table_recovery_status}
            LOGGER.debug(f'unassigning changelog partitions: {[p for p in unassign]}')
            self.consumer.incremental_unassign(list(unassign.values()))
            for p in unassign:
                del self.consumer.temp_assign_recovery_partitions[p]
            # for p_obj in self.consumer.temp_assign_recovery_partitions:
            #     del self.consumer.table_recovery_status[p_obj.partition]
            LOGGER.info("TABLE RECOVERY COMPLETE!")
        LOGGER.debug(f'ASSIGNMENT BEFORE RESUME:\n{[p_obj.partition for p_obj in self.consumer.assignment()]}')
        # self.consumer.resume(self.consumer.temp_assign_partitions)
        resume = [p_obj for p_obj in self.consumer.assignment() if '__changelog' not in p_obj.topic]
        self.consumer.resume(resume)
        LOGGER.debug(f'Resuming consumption for partitions:\n{[p_obj.partition for p_obj in resume]}')
        LOGGER.info('Continuing normal consumption loop...')
    
    def run(self, *args, **kwargs):
        transaction = None
        try:
            while True:
                try:
                    transaction = TableTransaction(self.consume(), self.producer, self.consumer, self.metrics_manager, self.changelog_topic)
                    self.app_function(transaction, *args, **kwargs)
                    if not transaction._committed:
                        transaction.commit()
                except NoMessageError:
                    self.producer.poll(0)
                    LOGGER.debug('No messages!')
                except RunTableRecovery:
                    self._table_recovery()
                except Exception as e:
                    raise
        finally:
            LOGGER.info('App is shutting down...')
            try:
                if transaction._active_transaction:
                    self.producer.abort_transaction(10)
            except:
                pass
            finally:
                shutdown_cleanup(consumer=self.consumer)
                _rdb_close()



# ---------------------------------------------------------
#
# Globally accessed methods necessary for table management
#
# ---------------------------------------------------------


def _rdb_close(partitions=None):
    if not partitions:
        partitions = list(rdb_tables.keys())
    LOGGER.debug(f'RocksDB - closing connections for partitions {partitions}')
    for k in partitions:
        rdb_tables[k].close()
        del rdb_tables[k]
        LOGGER.debug(f'partition {k} RocksDB connection closed.')
    LOGGER.info(f'RocksDB - closed connections for partitions {partitions}')


def _rdb_init(partition):
    rdb_tables[partition] = RDB(f'p{partition}')
    LOGGER.debug(f'RocksDB connection for partition {partition} initialized')


def _rocksdb_assign(consumer, partition_objs):
    LOGGER.debug(f'Consumer - Assigning additional partitions: {[p_obj.partition for p_obj in partition_objs]}')
    consumer.temp_assign_partitions = partition_objs
    if partition_objs:
        _init_rdb_tables(partition_objs)
        changelog_partitions = _changelog_partition_dict(partition_objs, consumer.changelog_topic)
        tables_to_recover = _table_recovery_offsets(rdb_tables, _get_changelog_highwaters(consumer, changelog_partitions))
        consumer.table_recovery_status.update(tables_to_recover)
        consumer.incremental_assign(partition_objs)
        if consumer.table_recovery_status:
            # TODO: This might be able to be managed earlier on as part of the "table recovery" logic.
            consumer.pause(partition_objs)
            changelog_partitions = _changelog_current_offsets(changelog_partitions, tables_to_recover)
            consumer.temp_assign_recovery_partitions.update(changelog_partitions)
            LOGGER.info(f'Table recovery required: {consumer.table_recovery_status}')
            consumer.incremental_assign(list(changelog_partitions.values()))
    else:
        LOGGER.info('No new/additional partition assignments required!')
        LOGGER.info(f'Resuming current assignment of: {[p_obj.partition for p_obj in consumer.assignment()]}')
    LOGGER.info('Consumer - partition assignment complete.')
    if consumer.table_recovery_status:
        # Due to confluent-kafka consumers asynchronous operations, the best way to force a desired
        # control flow within the general app consumer loop is to raise an exception and
        # interrupt whatever the app is currently doing and start recovering.
        raise RunTableRecovery

def _rocksdb_unassign(consumer, partition_objs):
    partitions = [p_obj.partition for p_obj in partition_objs]
    LOGGER.debug(f'Consumer - Unassigning topic {consumer.topic} partitions: {partitions}')
    consumer.incremental_unassign(partition_objs)
    for p in partitions:
        if p in consumer.table_recovery_status:
            del consumer.table_recovery_status[p]
        if p in consumer.temp_assign_recovery_partitions:
            del consumer.temp_assign_recovery_partitions[p]
    LOGGER.debug(f'table_recovery_status after unassignment: {consumer.table_recovery_status}')
    LOGGER.debug(f'temp_assign_recovery_partitions after unassignment: {consumer.temp_assign_recovery_partitions}')
    _rdb_close(partitions)
    LOGGER.info('Consumer - Unassignment request complete.')


def _init_rdb_tables(partition_objs):
    for p in partition_objs:
        partition = p.partition
        if partition not in rdb_tables:
            _rdb_init(partition)


def _changelog_partition_dict(partition_objs, changelog_topic):
    return {p_obj.partition: TopicPartition(topic=changelog_topic, partition=p_obj.partition, offset=p_obj.offset) for p_obj in partition_objs}


def _get_changelog_highwaters(consumer, changelog_partitions):
    return {p: consumer.get_watermark_offsets(p_obj) for p, p_obj in changelog_partitions.items()}


def _table_recovery_offsets(rdb_tables, changelog_watermarks):
    tables_to_recover = {}
    for partition, watermarks in changelog_watermarks.items():
        LOGGER.debug(f'(lowwater, highwater) for changelog p{partition}: {watermarks}')
        table_offset = rdb_tables[partition].read('offset')
        if table_offset:
            table_offset = int(table_offset)
        else:
            table_offset = 0
        if table_offset < watermarks[1]:
            tables_to_recover[partition] = {'table_offset': table_offset, 'watermarks': watermarks}
    return tables_to_recover


def _changelog_current_offsets(changelog_partitions, tables_to_recover):
    partitions_out = {}
    topic = changelog_partitions[0].topic
    for p, offsets in tables_to_recover.items():
        new_offset = tables_to_recover[p]['table_offset']
        low_mark = tables_to_recover[p]['watermarks'][0]
        if new_offset < low_mark:
            LOGGER.debug(f'p{p} changelog offset is short by {low_mark - new_offset} commits')
            new_offset = low_mark
        changelog_partitions[p].offset = new_offset
        partitions_out[p] = TopicPartition(topic=topic, partition=p, offset=new_offset)
    return partitions_out



import json
import logging
from os import environ

import pytest

LOGGER = logging.getLogger(__name__)


@pytest.fixture()
def actual_results():
    LOGGER.info('Preparing actual results...')
    with open(environ['TEST_DATA_OUT'], 'r') as f:
        results = json.load(f)
    return [{k: v for k, v in result.items() if k in ['key', 'value']} for result in results]


@pytest.fixture()
def expected_results(map_expected_output):
    LOGGER.info('Preparing expected results...')
    return [{k: v for k, v in message.items() if k in ['key', 'value']} for message in map_expected_output]


def validate_results(actual_results, expected_results):
    LOGGER.info('Validating results...')
    try:
        for expected, actual in zip(expected_results, actual_results):
            for k, v in expected['value'].items():
                assert expected['value'][k] == actual['value'][k]
        LOGGER.info('TEST SUCCESS!')
    except AssertionError:
        LOGGER.info(
            f"TEST FAILURE =(\n\nEXPECTED=\n{json.dumps(expected_results, indent=4)}\n\nACTUAL:\n{json.dumps(actual_results, indent=4)}")
        raise



import json
import logging
import subprocess
from os import environ, remove
from time import sleep
from unittest.mock import patch

import psutil
import pytest

LOGGER = logging.getLogger(__name__)


@pytest.fixture()
def setup_teardown_env_vars(env_vars):
    LOGGER.info("Patching environment variables...")
    env_patch = patch.dict('os.environ', env_vars)
    env_patch.start()
    yield None
    LOGGER.info("Unpatching environment variables...")
    env_patch.stop()


@pytest.fixture()
def setup_teardown_app():
    LOGGER.info("Initializing app...")
    parent = subprocess.Popen("dude app run", shell=True)
    sleep(30)  # wait for app to launch
    yield None
    LOGGER.info("Terminating app...")
    children = psutil.Process(parent.pid)
    for child in children.children(recursive=True):
        child.terminate()


@pytest.fixture()
def setup_teardown_test_data(map_integration_input):
    LOGGER.info("Creating test data input file...")
    with open(environ['TEST_DATA_IN'], 'w') as f:
        json.dump(map_integration_input, f, indent=4, sort_keys=True)
    yield None
    # LOGGER.info("Deleting test data files...")
    remove(environ['TEST_DATA_IN'])
    remove(environ['TEST_DATA_OUT'])



import ssl
import logging
from logging.handlers import RotatingFileHandler
from .faust_runtime_vars import env_vars
from faust import SASLCredentials
from nubium_topic_configs.clusters import cluster_topic_map
from nubium_topic_configs.topics import primary_topics

LOGGER = logging.getLogger(__name__)


def get_ssl_context():
    """
    Constructs SASL context based on environment variables
    """
    if env_vars()["RHOSAK_USERNAME"]:
        LOGGER.info('Using SASL-authenticated producers/consumers!')
        return {
            'broker_credentials': SASLCredentials(
                username=env_vars()['RHOSAK_USERNAME'],
                password=env_vars()['RHOSAK_PASSWORD'],
                ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH))
        }
    else:
        LOGGER.info('Authentication/encryption disabled for producers/consumers!')
        return {}


def get_data_store():
    try:
        store = env_vars()['STORE']
    except KeyError:
        if env_vars()['USE_ROCKSDB'].lower() == 'true':
            store = 'rocksdb://'
        else:
            store = 'memory://'
    return {'store': store}


def get_file_logging_handler():
    if env_vars()['KIWF_LOG_FILEPATH']:
        handler = RotatingFileHandler(env_vars()['KIWF_LOG_FILEPATH'], maxBytes=100000, backupCount=1)
        handler.setLevel(env_vars()['LOGLEVEL'])
        handler.setFormatter(logging.Formatter('[%(asctime)s] [%(process)d] [%(levelname)s] %(message)s'))
        return {'loghandlers': [handler]}
    return {}


def get_config():
    """
    Maps environment variables to Faust app config dictionary
    Any environ.get config options are generally considered optional and generaly default to .
    :return: Faust configs
    :rtype: dict
    """
    broker_timeout = int(env_vars()['CONSUMER_HEARTBEAT_TIMEOUT_SECONDS'])
    config = {
        'id': env_vars()['APP_NAME'],
        'broker': cluster_topic_map()[env_vars()['BROKER_TOPIC_NAME']],
        'topic_partitions': int(env_vars()['TOPIC_PARTITIONS']) if env_vars()['TOPIC_PARTITIONS'] else primary_topics[env_vars()['BROKER_TOPIC_NAME'].replace("__TEST", "")].partitions,
        'topic_replication_factor': int(env_vars()['TOPIC_REPLICATION_FACTOR']),
        'processing_guarantee': env_vars()['PROCESSING_GUARANTEE'],
        'stream_recovery_delay': int(env_vars()['STREAM_RECOVERY_DELAY']),
        'consumer_auto_offset_reset': env_vars()['CONSUMER_AUTO_OFFSET_RESET'],
        'broker_max_poll_records': int(env_vars()['BROKER_MAX_POLL_RECORDS']),
        'broker_session_timeout': broker_timeout,
        'broker_request_timeout': broker_timeout * 2,
        'datadir': f"{env_vars()['FAUST_DATADIR_BASE']}/{env_vars()['HOSTNAME']}/"
    }

    config.update(get_data_store())
    config.update(get_ssl_context())
    config.update(get_file_logging_handler())
    return config



import logging
import json
from nubium_utils import parse_headers
from nubium_utils.custom_exceptions import ProduceHeadersException, MaxRetriesReached, RetryTopicSend, FailureTopicSend
from nubium_utils.metrics import MetricsManager
from .message_utils import confirm_produce
from .confluent_runtime_vars import env_vars
from .confluent_configs import get_kafka_configs, init_producer_configs, init_transactional_producer_configs, producer_serializers
from confluent_kafka.avro import AvroProducer, loads
from confluent_kafka.schema_registry.avro import AvroSerializer
from confluent_kafka import SerializingProducer

LOGGER = logging.getLogger(__name__)


def get_producers(topic_schema_dict, return_as_singular=True, cluster=None):
    topic_dict = {}
    LOGGER.debug('Setting up Kafka Producer')
    for topic, schema in topic_schema_dict.items():
        producer = AvroProducer(
            init_producer_configs(topic, *get_kafka_configs(), cluster),
            default_key_schema=loads("""{"type": "string"}"""),
            default_value_schema=loads(json.dumps(schema)))
        LOGGER.debug('Producer setup complete.')
        setattr(producer, 'topic', topic)
        topic_dict[topic] = producer
    if return_as_singular:
        if len(topic_dict) == 1:
            return list(topic_dict.values())[0]
    return topic_dict


def get_transactional_producer(topic_schema_dict, schema_registry, cluster=None):
    LOGGER.debug('Setting up Kafka Transactional Producer')
    producer = SerializingProducer(init_transactional_producer_configs(topic_schema_dict, schema_registry, get_kafka_configs()[0], cluster))
    LOGGER.debug('Producer setup complete.')
    setattr(producer, 'schema_dict', producer_serializers(topic_schema_dict, schema_registry))
    producer.init_transactions()
    return producer


def produce_message(producer, producer_kwargs: dict, metrics_manager: MetricsManager = None,
                    consumed_msg_headers_passthrough: tuple = ()):
    """
    Helper for producing a message with confluent_kafka producers; primarily handles headers passthrough and enforcing
    all required headers fields exist before producing.
    You can pass in the previous message headers via 'consumed_msg_headers_passthrough', which extracts the headers.
    Then, if you wish to overwrite any of them, you can provide your own "headers" keyword in 'producer_kwargs'
    which will take that dict and overwrite any matching key in the 'consumed_msg_headers_passthrough' argument.
    You can also provide a None value for a key to remove it entirely.
    :param producer: A confluent_kafka producer
    :type: confluent_kafka.Producer or confluent_kafka.avro.AvroProducer
    :param producer_kwargs: the kwargs to pass into the producer instance
    :type producer_kwargs: dict
    :param metrics_manager: a MetricsManager object instance
    :type metrics_manager: MetricsManager
    :param consumed_msg_headers_passthrough: confluent_kafka Message.headers() from a consumed msg; will add all to new message
    :type consumed_msg_headers_passthrough: tuple
    """
    required_fields = ['guid', 'last_updated_by']

    if 'header' in producer_kwargs:
        raise KeyError('"header" is not the appropriate key for producing message headers; use "headers" instead')

    headers_out = parse_headers(consumed_msg_headers_passthrough)
    headers_out.update(producer_kwargs.get('headers', {}))
    headers_out = {key: value for key, value in headers_out.items() if value is not None}
    missing_required_keys = [key for key in required_fields if key not in headers_out]
    if missing_required_keys:
        raise ProduceHeadersException(f'Message headers are missing the required key(s): {missing_required_keys}')
    producer_kwargs['headers'] = headers_out

    LOGGER.debug('Polling for previous successful produce callbacks before adding additional produces to queue...')
    producer.poll(0)

    # multi-producer management
    if "topic" not in producer_kwargs and hasattr(producer, 'topic'):
        producer_kwargs.update({'topic': producer.topic})

    # single serialized producer management
    if hasattr(producer, 'schema_dict'):
        producer._value_serializer = producer.schema_dict[producer_kwargs['topic']]

    LOGGER.debug(f'Adding a message to the produce queue for topic {producer_kwargs["topic"]}')
    producer.produce(**producer_kwargs)
    if metrics_manager:
        metrics_manager.inc_messages_produced(1, producer_kwargs['topic'])
    LOGGER.info(f'Message added to the produce queue; GUID {headers_out.get("guid")}')


def produce_retry_message(message, producers, exception=None, value_schema=None):
    """
    Produces a message to the correct retry topic with an updated header
    Note: "exception" can be an Exception or plain str.

    The message header should contain a `kafka_retry_count` field,
    which is an integer representation of how many times the message has
    been tried for the current topic.
    If greater than the allowed maximum, produces to the retry topic.
    If less than the allowed maximum, produces to the current topic.
    """

    headers = dict(message.headers())
    guid = headers.get('guid', b'N/A').decode()
    kafka_retry_count = int(headers.get('kafka_retry_count', '0'))
    retry_topic = None

    if kafka_retry_count < int(env_vars()['RETRY_COUNT_MAX']):
        headers['kafka_retry_count'] = str(kafka_retry_count + 1)
        retry_topic = env_vars()['CONSUME_TOPICS']
    else:
        headers['kafka_retry_count'] = '0'
        retry_topic = env_vars().get('PRODUCE_RETRY_TOPICS', '')

    if isinstance(producers, dict):
        producer = producers.get(retry_topic, producers[env_vars()['CONSUME_TOPICS']])
    else:
        producer = producers
    if not value_schema:
        value_schema = producer._value_schema

    if retry_topic:
        if not exception:
            exception = RetryTopicSend()
        LOGGER.warning('; '.join([str(exception), f'retrying GUID {guid}']))
        LOGGER.debug('Polling for previous successful produce callbacks before adding additional produces to queue...')
        producer.poll(0)
        producer.produce(
            topic=retry_topic,
            value=message.value(),
            key=message.key(),
            headers=headers,
            value_schema=value_schema)
        confirm_produce(producer)
    else:
        if not exception:
            exception = FailureTopicSend()
        LOGGER.error('; '.join([str(exception), f'GUID {guid}']))
        produce_failure_message(message, producers, exception=MaxRetriesReached())


def produce_failure_message(message, producers, exception=None, value_schema=None):
    """
    Produces a message onto a deadletter queue.
    Note: "exception" can be an Exception or plain str.
    """
    headers = dict(message.headers())
    guid = headers.get('guid', b'N/A').decode()
    headers['kafka_retry_count'] = '0'
    failure_topic = env_vars()['PRODUCE_FAILURE_TOPICS']

    if isinstance(producers, dict):
        producer = producers.get(failure_topic, producers[env_vars()['CONSUME_TOPICS']])
    else:
        producer = producers

    if not value_schema:
        value_schema = producer._value_schema

    if not exception:
        exception = FailureTopicSend()
    LOGGER.error('; '.join([type(exception).__name__, str(exception), f'failing GUID {guid}']))
    headers["exception"] = json.dumps({"name": type(exception).__name__, "description": str(exception)})

    LOGGER.debug('Polling for previous successful produce callbacks before adding additional produces to queue...')
    producer.poll(0)
    LOGGER.debug(f'Adding a message to the produce queue for deadletter/failure topic {env_vars()["PRODUCE_FAILURE_TOPICS"]}')
    producer.produce(
        topic=failure_topic,
        value=message.value(),
        key=message.key(),
        headers=headers,
        value_schema=value_schema)
    LOGGER.info(f'Message added to the deadletter/failure topic produce queue; GUID {guid}')



from .consumer_utils import consume_message, consume_message_batch, enqueue_auto_commits, get_consumer
from .message_utils import success_headers, synchronous_message_handling, handle_no_messages, shutdown_cleanup
from .producer_utils import produce_message, produce_retry_message, produce_failure_message, get_producers
from .confluent_configs import init_metrics_pushing
from .dev_toolbox import KafkaToolbox
from .transaction_utils import TransactionApp, TableApp


import logging
from .message_utils import produce_message_callback, consume_message_callback
from nubium_utils.metrics import start_pushing_metrics
from .confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import WrappedSignals
from nubium_topic_configs.clusters import cluster_topic_map
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroSerializer, AvroDeserializer
import json
def patched_schema_loads(schema_str):
    from confluent_kafka.schema_registry.avro import Schema
    schema_str = schema_str.strip()
    if schema_str[0] != "{" and schema_str[0] != "[":
        schema_str = '{"type":' + schema_str + '}'

    return Schema(schema_str, schema_type='AVRO')
import confluent_kafka
confluent_kafka.schema_registry.avro._schema_loads = patched_schema_loads


LOGGER = logging.getLogger(__name__)
wrapped_signals = WrappedSignals()  # WrappedSignals needs to be initialized (somewhere) to catch signals


def init_ssl_configs():
    if env_vars()["RHOSAK_USERNAME"]:
        LOGGER.info('Using SASL-authenticated producers/consumers!')
        return {
            "security.protocol": "sasl_ssl",
            "sasl.mechanisms": "PLAIN",
            "sasl.username": env_vars()["RHOSAK_USERNAME"],
            "sasl.password": env_vars()["RHOSAK_PASSWORD"]
        }
    else:
        LOGGER.info('Authentication/encryption disabled for producers/consumers!')
    return {}


def init_schema_registry_configs(as_registry_object=False):
    """
    Provides the avro schema config
    :return: dict, None
    """
    config = None
    if as_registry_object:
        url = 'url'
    else:
        url = 'schema.registry.url'
    LOGGER.info(f'Using the schema server at {env_vars()["SCHEMA_REGISTRY_URL"]}')
    # RHOSAK registry
    if env_vars()['SCHEMA_REGISTRY_USERNAME']:
        config = {url: f"https://{env_vars()['SCHEMA_REGISTRY_SERVER']}"}
        LOGGER.debug(f'Using RHOSAK schema registry config: {config}')
    # local to Bifrost Nubium
    elif env_vars()['SCHEMA_REGISTRY_SSL_CA_LOCATION']:
        config = {
            url: f"https://{env_vars()['SCHEMA_REGISTRY_SERVER']}",
            'ssl.ca.location': env_vars()['SCHEMA_REGISTRY_SSL_CA_LOCATION'],
        }
    # Bifrost Nubium to Bifrost Nubium, local to local
    else:
        config = {url: f"http://{env_vars()['SCHEMA_REGISTRY_SERVER']}"}

    if as_registry_object:
        return SchemaRegistryClient(config)
    else:
        return config


def init_producer_configs(topic, ssl_configs=None, schema_registry_configs=None, cluster=None):
    """
    Sets up a single producer.
    """
    if not ssl_configs:
        ssl_configs = {}
    if not schema_registry_configs:
        schema_registry_configs = {}
    return {
        **ssl_configs,
        **schema_registry_configs,
        "bootstrap.servers": cluster if cluster else cluster_topic_map()[topic],
        "on_delivery": produce_message_callback,
        "acks": "all"}


def producer_serializers(schema_dict, schema_registry):
    return {topic: AvroSerializer(schema_registry, json.dumps(schema)) for topic, schema in schema_dict.items()}


def init_transactional_producer_configs(topic_schema_dict, schema_registry, ssl_configs=None, cluster=None):
    """
    Sets up a single producer.
    """
    if not cluster:
        check_topics = set([topic for topic in topic_schema_dict if topic in cluster_topic_map()])
        unmapped = set(topic_schema_dict.keys()) - check_topics
        if unmapped:
            LOGGER.warning(f'The topic(s) {unmapped} in the provided producer schema dict mapping are not included in'
                           f' the cluster topic map. There are valid use cases for this, such as dynamically adding'
                           f' new topics at runtime with a predetermined schema, or even just testing a new topic on'
                           f' the fly. Either way, make sure this is not a mistake!')
        cluster = set([cluster_topic_map()[topic] for topic in check_topics])
        if len(cluster) == 1:
            cluster = list(cluster)[0]
        else:
            raise ValueError('Cluster count for provided consumer topics did not equal 1. '
                             f'Please confirm your topics are on the same cluster. Topics:\n{topics}')
    if not ssl_configs:
        ssl_configs = {}
    return {
        **ssl_configs,
        "bootstrap.servers": cluster if cluster else cluster_topic_map()[topic],
        "on_delivery": produce_message_callback,
        "key.serializer": AvroSerializer(schema_registry, schema_str="""{"type": "string"}"""),
        "value.serializer": lambda: None,
        "enable.idempotence": "true",
        "transactional.id": env_vars()['HOSTNAME'],
        "acks": "all"}


def init_consumer_configs(topics, ssl_configs=None, schema_registry_configs=None, cluster=None):
    """
    Assummes topics are all in the same cluster.
    """
    if not cluster:
        cluster = set([cluster_topic_map()[topic] for topic in topics])
        if len(cluster) == 1:
            cluster = list(cluster)[0]
        else:
            raise ValueError('Cluster count for provided consumer topics did not equal 1. '
                             f'Please confirm your topics are on the same cluster. Topics:\n{topics}')
    if not ssl_configs:
        ssl_configs = {}
    if not schema_registry_configs:
        schema_registry_configs = {}
    max_time_between_consumes_mins = int(env_vars()['CONSUMER_TIMEOUT_LIMIT_MINUTES']) + int(env_vars()['TIMESTAMP_OFFSET_MINUTES'])
    return {
        **ssl_configs,
        **schema_registry_configs,
        "bootstrap.servers": cluster,
        "group.id": env_vars()['APP_NAME'],
        "group.instance.id": env_vars()['HOSTNAME'],
        "on_commit": consume_message_callback,
        "enable.auto.commit": True if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT']=='true' else False,
        "auto.offset.reset":  'earliest' if env_vars()['CONSUMER_AUTO_OFFSET_RESET']=='earliest' else 'latest',
        "auto.commit.interval.ms": int(env_vars()['CONSUMER_AUTO_COMMIT_INTERVAL_SECONDS']),
        "enable.auto.offset.store": False if env_vars()['CONSUMER_ENABLE_AUTO_COMMIT']=='true' else True,
        "max.poll.interval.ms": 60000 * max_time_between_consumes_mins,
        "session.timeout.ms": int(env_vars()['CONSUMER_HEARTBEAT_TIMEOUT_SECONDS']) * 1000,
        "message.max.bytes": int(env_vars()['MESSAGE_BATCH_MAX_MB']),
        "fetch.max.bytes": int(env_vars()['MESSAGE_TOTAL_MAX_MB']),
        "queued.max.messages.kbytes": int(env_vars()['MESSAGE_QUEUE_MAX_MB'])}


def init_transactional_consumer_configs(topics, schema_registry, ssl_configs=None, cluster=None, schema=None):
    """
    Assummes topics are all in the same cluster.
    """
    if not cluster:
        cluster = set([cluster_topic_map()[topic] for topic in topics])
        if len(cluster) == 1:
            cluster = list(cluster)[0]
        else:
            raise ValueError('Cluster count for provided consumer topics did not equal 1. '
                             f'Please confirm your topics are on the same cluster. Topics:\n{topics}')
    if not ssl_configs:
        ssl_configs = {}
    max_time_between_consumes_mins = int(env_vars()['CONSUMER_TIMEOUT_LIMIT_MINUTES']) + int(env_vars()['TIMESTAMP_OFFSET_MINUTES'])
    # with patch('confluent_kafka.schema_registry.avro._schema_loads', side_effect=patched_schema_loads):
    #     from confluent_kafka.schema_registry.avro import AvroDeserializer

    return {
        # Connection Settings
        **ssl_configs,
        "bootstrap.servers": cluster,

        # General Consumer Settings
        "group.id": env_vars()['APP_NAME'],
        # "group.instance.id": env_vars()['HOSTNAME'],
        "on_commit": consume_message_callback,
        "auto.offset.reset":  'earliest' if env_vars()['CONSUMER_AUTO_OFFSET_RESET']=='earliest' else 'latest',

        # Transaction Settings
        "isolation.level": "read_committed",
        "enable.auto.commit": False,
        "enable.auto.offset.store": False,

        # Registry Serialization Settings
        "key.deserializer": AvroDeserializer(schema_registry, schema_str='{"type": "string"}'),
        "value.deserializer": AvroDeserializer(schema_registry, schema_str=json.dumps(schema) if schema else None),

        # Performance Settings
        "partition.assignment.strategy": 'cooperative-sticky',
        "max.poll.interval.ms": 60000 * max_time_between_consumes_mins,
        "session.timeout.ms": int(env_vars()['CONSUMER_HEARTBEAT_TIMEOUT_SECONDS']) * 1000,
        "message.max.bytes": int(env_vars()['MESSAGE_BATCH_MAX_MB']),
        "fetch.max.bytes": int(env_vars()['MESSAGE_TOTAL_MAX_MB']),
        "queued.max.messages.kbytes": int(env_vars()['MESSAGE_QUEUE_MAX_MB']),
    }


def init_metrics_pushing(metrics_manager):
    if env_vars()['DO_METRICS_PUSHING'].lower() == 'true':
        LOGGER.info('Metric pushing to gateway ENABLED')
        start_pushing_metrics(metrics_manager, int(env_vars()['METRICS_PUSH_RATE']))
    else:
        LOGGER.info('Metric pushing to gateway DISABLED')


def get_kafka_configs():
    return init_ssl_configs(), init_schema_registry_configs()
